
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

use work.const.all;

entity dcache is
port(clk, reset: in std_logic;
	mem, rd, wr: in std_logic;
	sel8, sel16, sel32: in std_logic;
	addr: in adr_type;
	data: out dat_type;
	tlb_addr: in adr_type;

	buf_dat: in dat_type;
	buf_mask: in std_logic_vector(3 downto 0);
	buf_empty: in std_logic;
	o_drain: out std_logic;

	mem_iline: in line_dat_type;
	mem_oline: out line_dat_type;
	o_miss: out std_logic;
	memfill: in std_logic;
	mem_addr: out adr_type;
	mem_rd, mem_wr: out std_logic;
	excep: out std_logic;
	tlb_miss: in std_logic
);
end entity dcache;

architecture rtl of dcache is
	subtype byte_type is std_logic_vector(7 downto 0);
	type byte_ary_type is array (0 to 3) of byte_type;
	type cache_dat_type is array (0 to 3) of line_dat_type;
	type cache_tag_type is array (0 to 3) of line_tag_type;
	type cache_valid_type is array (0 to 3) of std_logic;

	signal ary_dirty: cache_valid_type;
	signal ary_valid: cache_valid_type;
	signal ary_tag: cache_tag_type;
	signal ary_data: cache_dat_type;

	signal dirty, miss, valid, fill: std_logic;
	signal store_line, load_line, cache_line : line_dat_type;
	signal m_tag, itag, tag: line_tag_type;
	signal ncol, nrow: Natural;
	signal row: line_row_type;
	signal col: line_col_type;
	signal seldata, seldirty, sel: std_logic_vector(3 downto 0);
	signal b8e: std_logic_vector(3 downto 0);
	signal b16e: std_logic_vector(1 downto 0);
	signal ad0, ad1: std_logic;
	signal store, byte, odata, oword: byte_ary_type;
	signal store_word: dat_type;
	signal clean, hit, drain, s_drain: std_logic;

	signal m_rd, m_wr: std_logic;
begin
	ad0 <= addr(0); ad1 <= addr(1);

	excep <= ((sel32 and (ad0 or ad1)) or (sel16 and ad0)) and mem;
	row <= addr(line_row_b downto line_row_e);
	col <= addr(line_col_b downto line_col_e);
	nrow <= conv_integer(row);
	ncol <= conv_integer(col);
	itag <= tlb_addr(line_tag_b downto line_tag_e);

	valid <= ary_valid(nrow);
	dirty <= ary_dirty(nrow);
	tag <= ary_tag(nrow);
	cache_line <= ary_data(nrow);

	-- byte handling stuff
	b16e(0) <= sel16 and not ad1;
	b16e(1) <= sel16 and ad1;
	b8e(0) <= sel8 and not ad1 and not ad0; -- sel8 and "00"
	b8e(1) <= sel8 and not ad1 and ad0;     -- sel8 and "01"
	b8e(2) <= sel8 and ad1 and not ad0;     -- sel8 and "10"
	b8e(3) <= sel8 and ad1 and ad0;         -- sel8 and "11"

	byte(0) <= cache_line(ncol)(7 downto 0);
	byte(1) <= cache_line(ncol)(15 downto 8);
	byte(2) <= cache_line(ncol)(23 downto 16);
	byte(3) <= cache_line(ncol)(31 downto 24);

	store(0) <= buf_dat(7 downto 0);
	store(1) <= buf_dat(15 downto 8);
	store(2) <= buf_dat(23 downto 16);
	store(3) <= buf_dat(31 downto 24);

	defdat: for i in 0 to 3 generate
		oword(i) <= store(i) when buf_mask(i) = '1' else
		            byte(i);
	end generate;


	odata(0) <= oword(0) when b8e(0) = '1' or b16e(0) = '1' or sel32 = '1' else
	            oword(1) when b8e(1) = '1' else
	            oword(2) when b8e(2) = '1' or b16e(1) = '1' else
	            oword(3) when b8e(3) = '1' else
	            (others => '0');
	odata(1) <= oword(1) when b16e(0) = '1' or sel32 = '1' else
	            oword(3) when b16e(1) = '1' else
	            (others => '0');
	odata(2) <= oword(2) when sel32 = '1' else
	            (others => '0');
	odata(3) <= oword(3) when sel32 = '1' else
	            (others => '0');

	-- miss and fill stuff
	defdrain:process(clk, reset)
	begin
		-- miss is a combination signal which depends
		-- of the input address which is modified in
		-- a miss where the store buffer writes in the
		-- store buffer. So this register tries to hold
		-- the value.
		if reset = '1' then
			s_drain <= '0';
		elsif clk = '1' and clk'event then
			s_drain <= drain;
		end if;
	end process;

	hit <= 	'1' when valid = '1' and tag = itag else
	        '0';
	miss <= (not hit or s_drain) and mem and not tlb_miss;
	drain <= miss and not buf_empty;
	fill <= miss and memfill;
	o_miss <= miss;
	-- It is needed a cycle before draining, because in other case
	-- we will have a loop. Once that s_drain holds a 1, then we
	-- can modify the output drain signal and miss will remain
	-- to 1 until the new line is in our cache
	o_drain <= s_drain and not buf_empty;

	-- signals for main memory
	clean <= dirty and valid;
	m_rd <= miss and not clean and not drain;
	m_wr <= miss and clean and not drain;
	m_tag <= tag when m_wr = '1' else itag;
	mem_wr <= m_wr;
	mem_rd <= m_rd;
	mem_addr(line_tag_b downto line_tag_e) <= m_tag;
	mem_addr(line_row_b downto line_row_e) <= row;
	mem_addr(line_col_b downto 0) <= (others => '0');

	-- cache memory
	store_word(7 downto 0) <= store(0) when buf_mask(0) = '1' else
	                          byte(0);
	store_word(15 downto 8) <= store(1) when buf_mask(1) = '1' else
	                           byte(1);
	store_word(23 downto 16) <= store(2) when buf_mask(2) = '1' else
	                            byte(2);
	store_word(31 downto 24) <= store(3) when buf_mask(3) = '1' else
	                            byte(3);

	mem_oline <= cache_line;
	defslin: for i in 0 to words_per_line-1 generate
		store_line(i) <= store_word when ncol = i and wr = '1' else
		                 cache_line(i) when ncol /=i and wr = '1' else
		                 mem_iline(i);
	end generate;

	defcach: for i in 0 to 3 generate
		sel(i) <=  '1' when nrow = i else '0';
		seldirty(i) <= sel(i) and (wr or (m_wr and fill));
		seldata(i) <= sel(i) and ((fill and not dirty) or wr);

		defdir: process(clk, reset)
		begin
			if reset = '1' then
				ary_dirty(i) <= '0';
			elsif clk = '1' and clk'event then
				if seldirty(i) = '1' then
					ary_dirty(i) <= wr;
				end if;
			end if;
		end process;

		defline: process(clk, reset, mem_iline, fill, sel(i))
		begin
			if reset = '1' then
				ary_valid(i) <= '0';
				ary_tag(i) <= (others => '0');
				ary_data(i) <= (others => (others => '0'));
			elsif clk = '1' and clk'event then
				if seldata(i) = '1' then
					ary_valid(i) <= '1';
					ary_tag(i) <= itag;
					ary_data(i) <= store_line;
				end if;
			end if;
		end process;
	end generate;

	data <= odata(3) & odata(2) & odata(1) & odata(0);
end architecture rtl;

