
library ieee;
use ieee.std_logic_1164.all;

use work.const.all;

entity storeb is
	PORT(clk, reset: in std_logic;
	     wr, shift: in std_logic;
	     sel8, sel16, sel32: in std_logic;
	     datin: in dat_type;
	     addrin: in dat_type;

	     cache_addr: out adr_type;
	     datout: out dat_type;
	     hit: out std_logic_vector(3 downto 0);

	     empty, full, cache_wr: out std_logic
);
end entity storeb;

architecture rtl of storeb is
	subtype buf_adr is std_logic_vector(adr_w-3 downto 0);
	subtype buf_mask is std_logic_vector(3 downto 0);
	subtype byte is std_logic_vector(7 downto 0);
	type buf_dat is array(0 to 3) of byte;
	type buf_dat_type is array (0 to 3) of buf_dat;
	type buf_adr_type is array (0 to 3) of buf_adr;
	type buf_mask_type is array (0 to 3) of buf_mask;
	type buf_valid_type is array (0 to 3) of std_logic;

	signal load_data, ary_data: buf_dat_type;
	signal load_addr, ary_addr: buf_adr_type;
	signal load_valid, ary_valid: buf_valid_type;
	signal load_mask, ary_mask: buf_mask_type;
	signal addr: buf_adr;
	signal store, idata, odata: buf_dat;
	signal sel, wr_ena, mask, free: std_logic_vector(3 downto 0);
	signal laddr: std_logic_vector(1 downto 0);
	signal o_hit: std_logic_vector(3 downto 0);
begin
	addr <= addrin(adr_w-1 downto 2);
	empty <= not ary_valid(0) and not ary_valid(1) and
		 not ary_valid(2) and not ary_valid(3);
	full <= ary_valid(0) and ary_valid(1) and
	       ary_valid(2) and ary_valid(3);
	cache_addr <= ary_addr(0) & "00";
	cache_wr <= shift and ary_valid(0);

	-- define output data
	defsel: for i in 0 to 3 generate
		-- sel(i) = 1 when data from position i is related to addr
		sel(i) <= '1' when ary_addr(i) = addr and
		                   ary_valid(i) = '1' else
		          '0';
	end generate;

	defodat: for i in 0 to 3 generate
		odata(i) <= ary_data(3)(i) when sel(3) = '1' and
		                               ary_mask(3)(i) = '1' else
		           ary_data(2)(i) when sel(2) = '1' and
		                               ary_mask(2)(i) = '1' else
		           ary_data(1)(i) when sel(1) = '1' and
		                               ary_mask(1)(i) = '1' else
		           ary_data(0)(i) when sel(0) = '1' and
		                               ary_mask(0)(i) = '1' else
		           (others => 'X');
		o_hit(i) <= ary_mask(0)(i) or ary_mask(1)(i) or
		            ary_mask(2)(i) or ary_mask(3)(i);
	end generate;

	hit <= o_hit when shift = '0' else ary_mask(0);
	datout(7 downto 0)   <= odata(0) when shift = '0' else ary_data(0)(0);
	datout(15 downto 8)  <= odata(1) when shift = '0' else ary_data(0)(1);
	datout(23 downto 16) <= odata(2) when shift = '0' else ary_data(0)(2);
	datout(31 downto 24) <= odata(3) when shift = '0' else ary_data(0)(3);

	-- define load of the registers
	idata(0) <= datin(7 downto 0);
	idata(1) <= datin(15 downto 8);
	idata(2) <= datin(23 downto 16);
	idata(3) <= datin(31 downto 24);

	store(0) <= (others => 'X') when mask(0) = '0' else
	            idata(0);
	store(1) <= (others => 'X') when mask(1) = '0' else
	            idata(0) when sel8 = '1' else
	            idata(1);
	store(2) <= (others => 'X') when mask(2) = '0' else
	            idata(0) when sel8 = '1' or sel16 = '1' else
	            idata(2);
	store(3) <= (others => 'X') when mask(3) = '0' else
	            idata(0) when sel8 = '1' else
	            idata(1) when sel16 = '1' else
	            idata(3);

	laddr <= addrin(1 downto 0);
	mask <= "0001" when sel8 = '1' and laddr = "00" else
	        "0010" when sel8 = '1' and laddr = "01" else
	        "0100" when sel8 = '1' and laddr = "10" else
	        "1000" when sel8 = '1' and laddr = "11" else
	        "0011" when sel16 = '1' and laddr(1) = '0' else
	        "1100" when sel16 = '1' and laddr(1) = '1' else
	        "1111" when sel32 = '1' else
	        "XXXX";

	defld: process(shift, wr, mask, store, addr,
	               ary_mask, ary_data, ary_addr, ary_valid)
	begin
		for i in 0 to 3 loop
			if shift = '1' then
				if i = 3 then
					load_mask(i) <= (others => '0');
					load_data(i) <= (others => (others => '0'));
					load_addr(i) <= (others => '0');
					load_valid(i) <= '0';
				else
					load_mask(i) <= ary_mask(i+1);
					load_data(i) <= ary_data(i+1);
					load_addr(i) <= ary_addr(i+1);
					load_valid(i) <= ary_valid(i+1);
				end if;
			else
				load_mask(i) <= mask;
				load_data(i) <= store;
				load_valid(i) <= '1';
				load_addr(i) <= addr;
			end if;
		end loop;
	end process;

	-- define the registers
	free <= "0001" when ary_valid(0) = '0' else
	        "0010" when ary_valid(1) = '0' else
	        "0100" when ary_valid(2) = '0' else
	        "1000" when ary_valid(3) = '0' else
	        "0000";

	defary: for i in 0 to 3 generate
		wr_ena(i) <= (free(i) and wr) or shift;

		process(clk, reset, wr_ena(i), load_data(i),
		        load_mask(i), load_addr(i), load_valid(i))
		begin
			if reset = '1' then
				ary_mask(i) <= (others => '0');
				ary_data(i) <= (others => (others => '0'));
				ary_addr(i) <= (others => '0');
				ary_valid(i) <= '0';
			elsif clk = '1' and clk'event and wr_ena(i) = '1' then
				ary_mask(i) <= load_mask(i);
				ary_data(i) <= load_data(i);
				ary_addr(i) <= load_addr(i);
				ary_valid(i) <= load_valid(i);
			end if;
		end process;
	end generate;
end architecture rtl;

