
#include <errno.h>
#include <setjmp.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "as.h"

extern struct label labels[MAXLABELS], *curlabel;
extern struct code codes[MAXCODES], *curcode;

static struct code *volatile codep;

static char names[][6] = {
	[HALT] = "HALT", [NOP] = "NOP",   [LDB] = "LDB",  [LDW] = "LDW",
	[LD] = "LD",     [STB] = "STB",   [STW] = "STW",  [ST] = "ST",
	[LDIH] = "LDIH", [LDIL] = "LDIL", [LDI] = "LDI",  [ADD] = "ADD",
	[SUB] = "SUB",   [OR] = "OR",     [AND] = "AND",  [JMP] = "JMP",
	[BRN] = "BRN",   [MOV] = "MOV",   [IRET] = "IRET",[IMOV] = "MOV",
	[FMOV] = "MOV",  [DATA] = "DATA", [FILL] = "FILL",[SLL] = "SLL",
	[SRL] = "SRL",   [XOR] = "XOR",   [NOT] = "NOT",  [NEG] = "NEG",
	[BS] = "BS",     [ITLB] = "ITLB", [DTLB] = "DTLB"
};

static int types [] = {
	[HALT] = 15,[NOP] = 0,   [LDB] = 0, [LDW] = 1,
	[LD] = 2,   [STB] = 0,   [STW] = 1, [ST] = 2,
	[LDIH] = 1, [LDIL] = 0,  [ADD] = 0, [SUB] = 1,
	[OR] = 2,   [AND] = 3,   [JMP] = 0, [BRN] = 1,
	[LDI] = 2,  [MOV] = 0,   [IRET] = 1,[IMOV] = 1,
	[FMOV] = 2, [DATA] = 0,  [SLL] = 5, [SRL] = 4,
	[XOR] = 6,  [NOT] = 7,   [NEG] = 1, [BS] = 0xf0,
	[ITLB] = 8, [DTLB] = 12
};

static void
error(char *fmt, ...)
{
	va_list va;
	va_start(va, fmt);
	error_(codep->line, fmt, va);
	va_end(va);
}

static char *
bin2asc(unsigned long b)
{
	static char buf[33], *cp;

	for (cp = buf; cp < &buf[32]; b <<= 1)
		*cp++ = ((b & 0x80000000) != 0) + '0';
	*cp = '\0';
	return buf;
}

static void
simple(struct code *code, int32_t imm)
{
	int op = code->op;
	unsigned long b = types[op] << 24;

	printf("%s\t;%08X %s\n", bin2asc(b), b, names[op]);
}

static void
load(struct code *code, int32_t imm)
{
	unsigned long b;
	int src1 = code->src1, src2 = code->src2, op = code->op;

	if (imm > 65536)
		error("address to far in load");
	b = (unsigned long) 0x1 << 28 |
	                    types[op] << 24 |
	                    src1 << 20 |
	                    src2 << 16 |
	                    (unsigned short) imm;

	printf("%s\t;%04X %s\t%04X(R%d),R%d\n",
	       bin2asc(b), b,
	       names[op], (unsigned short) imm, src1, src2);
}

static void
store(struct code *code, int32_t imm)
{
	unsigned long b;
	int src1 = code->src1, src2 = code->src2, op = code->op;

	if (imm > 65536)
		error("address to far in load");
	b = (unsigned long) 0x1 << 28 |
	                    0x1 << 26 |
	                    types[op] << 24 |
	                    src2 << 20 |
	                    src1 << 16 |
	                    (unsigned short) imm;

	printf("%s\t;%08X %s\tR%d,%04X(R%d)\n",
	       bin2asc(b), b,
	       names[op], src1,
	       (unsigned short) imm,  src2);
}

static void
ldi(struct code *code, int32_t imm)
{
	unsigned long b;
	int src1 = code->src1, op = code->op;

	if (imm > 65536)
		error("inmediate value too big in a ldi");
	b = (unsigned long) 0x2 << 28 |
	                    types[op] << 26 |
	                    src1 << 16 |
	                    (unsigned short) imm;

	printf("%s\t;%08X %s\t%04X,R%d\n",
	       bin2asc(b), b,
	       names[op], (unsigned short) imm, src1);
}

static void
alu(struct code *code, int32_t imm)
{
	unsigned long b;
	int src1 = code->src1, src2 = code->src2;
	int dst = code->dst, op = code->op, type = types[op];

	b = (unsigned long) 0x4 << 28 |
	                    (type&0x3) << 26 |
	                    (type&0xc) << 22 |
	                    src1 << 20 |
	                    src2 << 16 |
	                    dst  << 12;

	printf("%s\t;%08X %s\tR%d,R%d,R%d\n",
	       bin2asc(b), b,
	       names[op], src1, src2, dst);
}

static void
imov(struct code *code, int32_t imm)
{
	unsigned long b;
	int src = code->src1, op = code->op;

	b = (unsigned long) types[op] << 24 |
	                    src << 20;

	printf("%s\t;%08X %s\tR%d,I\n",
	       bin2asc(b), b,
	       names[op], src);
}

static void
fmov(struct code *code, int32_t imm)
{
	unsigned long b;
	int dst = code->dst, op = code->op;

	b = (unsigned long) types[op] << 24 |
	                    dst << 12;

	printf("%s\t;%08X %s\tF,R%d\n",
	       bin2asc(b), b,
	       names[op], dst);
}

static void
mov(struct code *code, int32_t imm)
{
	unsigned long b;
	int dst = code->dst, op = code->op, type = types[op];
	int src1 = code->src1, src_shift, dst_shift;

	src_shift = (op == NEG) ? 16 : 20;
	dst_shift = (op == ITLB || op == DTLB) ? 16 : 12;

	b = (unsigned long) 0x4 << 28 |
	                    (type&0x3) << 26 |
	                    (type&0xc) << 22 |
	                    src1 << src_shift |
	                    dst  << dst_shift;

	printf("%s\t;%08X %s\tR%d,R%d\n",
	       bin2asc(b), b,
	       names[op], src1, dst);
}

static void
jmp(struct code *code, int32_t imm)
{
	int op = code->op;
	unsigned long b;
	char *fmt;

	if (imm > 65536)
		error("incorrect inmediate value in jmp");
	b = (unsigned long) 0x8 << 28 |
	                    types[op] << 24 |
	                    imm >> 2 & 0xFFFF;

	fmt = (op == IRET) ? "%s\t;%08X %s\n" : "%s\t;%08X %s\t%04X\n";
	printf(fmt,
	       bin2asc(b), b,
	       names[op], (unsigned short) imm);
}

static void
brn(struct code *code, int32_t imm)
{
	static char cond[][3] = {
		"Z", "NZ", "C", "NC"
	};
	unsigned long b, dest;

	if (labs(imm) > 65536)
		error("incorrect inmediate value for brn '%ld'", (long) imm);
	b = (unsigned long) 0x8 << 28 |
	                    0x1 << 26 |
	                    code->cond << 24 |
	                    code->src1 << 20 |
	                    code->src2 << 16 |
	                    imm >> 2 & 0xFFFF;

	dest = (code - codes << 2) + imm;
	printf("%s\t;%08X %s\t%s,%04X,R%d,R%d -> %08X\n",
	       bin2asc(b), b,
	       names[code->op], cond[code->cond], (unsigned short) imm,
	       code->src1, code->src2, dest);
}

static void
data(struct code *code, int32_t imm)
{
	printf("%s\t;%s %08X\n",
	       bin2asc(imm), names[code->op], (unsigned long) imm);
}

void
pass2(void)
{
	static void (*ins[])(struct code *code, int32_t imm) = {
		[HALT] = simple,[NOP] = simple,[LDB] = load,  [LDW] = load,
		[LD] = load,    [STB] = store, [STW] = store, [ST] = store,
		[LDIH] = ldi,   [LDIL] = ldi,  [ADD] = alu,   [SUB] = alu,
		[OR] = alu,     [AND] = alu,   [JMP] = jmp,   [BRN] = brn,
		[LDI] = ldi,    [MOV] = mov,   [IRET] = jmp,  [IMOV] = imov,
		[FMOV] = fmov,  [DATA] = data, [FILL] = data, [SLL] = alu,
		[SRL] = alu,    [XOR] = alu,   [NOT] = mov,   [NOT] = mov,
		[NEG] = mov,    [BS] = simple, [ITLB] = mov,  [DTLB] = mov,
	};
	volatile unsigned pc;

	for (pc = 0, codep = codes; codep < curcode; pc += 4, ++codep) {
		struct label *lp;
		extern jmp_buf recovery;
		int32_t imm;

		if (setjmp(recovery))
			continue;
		if (codep->imm_islabel) {
			lp = &labels[codep->imm];
			if (!lp->def)
				error("label '%s' not defined", lp->s);
			if (codep->relative)
				imm = lp->offset - pc;
			else
				imm = lp->offset;
		} else {
			imm = codep->imm;
		}

		printf("%8.8X\t", pc);
		(*ins[codep->op])(codep, imm);
	}
}
