library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

use work.const.all;


entity tlb is
port (
	clk: in std_logic;
	reset: in std_logic;

	i_vadr: in adr_type;	-- virtual address (TLB 'converts' it to physical)
	wr, rd, ex: in std_logic; --what instruction wants to do in page
	i_data: in dat_type;	-- data from RX (tlb_write instruction)
	i_adr: in dat_type;		-- adr (tlb_write)
	wrEna: in std_logic;	-- whether to write i_data to TLB

	o_padr: out adr_type;	--physical address
	exc_miss, exc_access: out std_logic --exceptions (access should have bigger priority)
);
end entity tlb;

architecture behav of tlb is
	type tlb_virtual_type is array (0 to tlb_lines-1) of tlb_vadr_type;
	type tlb_physical_type is array (0 to tlb_lines-1) of tlb_padr_type;
	type tlb_status_type is array (0 to tlb_lines-1) of tlb_stat_type;

	signal ary_virt: tlb_virtual_type;
	signal ary_phys: tlb_physical_type;
	signal ary_stat: tlb_status_type;

	signal ld_virt, rd_virt, virt: tlb_vadr_type;
	signal ld_phys, rd_phys: tlb_padr_type;
	signal ld_stat, rd_stat: tlb_stat_type;

	signal wr_sel: std_logic_vector(tlb_lines-1 downto 0);
	signal hit: std_logic_vector(tlb_lines-1 downto 0);
	signal miss: std_logic;
	signal selEntry: std_logic_vector(tlb_entry_b-tlb_entry_e downto 0);
begin

	--parse input (tlb_write)
	ld_virt <= i_adr(adr_w-1 downto offset_w);
	ld_phys <= i_data(tlb_phys_b downto tlb_phys_e);
	ld_stat(v_pos) <= '1';
	ld_stat(d_pos) <= '0';
	ld_stat(r_pos downto x_pos) <= i_data(r_pos downto x_pos);

	--virtual address (input for 'conversion')
	virt <= i_vadr(adr_w-1 downto offset_w);

	tlbhit: process(ary_stat, virt)
	begin
		rd_virt <= (others => '0');
		rd_phys <= (others => '0');
		rd_stat <= (others => '0');
		hit <= (others => '0');

		for i in 0 to tlb_lines-1 loop
			if (ary_virt(i) = virt and ary_stat(i)(v_pos) = '1') then
				hit(i) <= '1';
				rd_virt <= ary_virt(i);
				rd_phys <= ary_phys(i);
				rd_stat <= ary_stat(i);
			end if;
		end loop;
	end process tlbhit;

	miss <= '1' when hit = "0000" else '0'; --find better way to represent "00"
	exc_miss <= miss;
	exc_access <=  '1' when ((rd = '1' and rd_stat(r_pos) = '0') or
							(wr = '1' and rd_stat(w_pos) = '0') or
							(ex = '1' and rd_stat(x_pos) = '0')) and
							miss = '0' else
					'0';

	selEntry <= i_data(tlb_entry_b downto tlb_entry_e);
	wrSel: process(selEntry)
	begin
		wr_sel <= (others => '0');
		for i in 1 to tlb_lines-1 loop
			if (conv_integer(selEntry) = i) then
				wr_sel(i) <= '1';
			end if;
		end loop;
	end process wrSel;

	tlbentries: for i in 0 to tlb_lines-1 generate
		tlbentry: process(clk, reset, wr_sel, wrEna)
		begin
			if reset = '1' then
				ary_virt(i) <= (others => '0');
				ary_phys(i) <= (others => '0');
				if i = 0 then
					ary_stat(i) <= "10111";
				else
					ary_stat(i) <= (others => '0');
				end if;
			elsif clk = '0' and clk'event and wr_sel(i) = '1' and wrEna = '1' then
				ary_virt(i) <= ld_virt;
				ary_phys(i) <= ld_phys;
				ary_stat(i) <= ld_stat;
			end if;
		end process;
	end generate;

	--o_padr(adr_w-1 downto tlb_phys_b+1) <= (others => '0');
	o_padr(tlb_phys_b downto 0) <= rd_phys & i_vadr(tlb_phys_e-1 downto 0);
	--o_padr <= i_vadr; --return pc, until tlb logic is finished
end architecture behav;
